'''Provides 'base' Nawah Core Module'''

from nawah.classes import Func, Perm
from nawah.utils import register_module

base = register_module(
    'base',
    desc='\'base\' module provides necessary structure to allow modules to access Base '
    'Functions Callables via \'call\' Utility',
    collection=None,
    attrs=None,
    funcs={
        'read': Func(
            permissions=[
                Perm(privilege='__sys'),
            ],
        ),
        'create': Func(
            permissions=[
                Perm(privilege='__sys'),
            ],
        ),
        'update': Func(
            permissions=[
                Perm(privilege='__sys'),
            ],
        ),
        'delete': Func(
            permissions=[
                Perm(privilege='__sys'),
            ],
        ),
    },
)



'''Provides 'Var' dataclass'''

from dataclasses import dataclass

from nawah.enums import VarType


@dataclass(kw_only=True)
class Var:
    '''Var dataclass serves role of placeholder variable value. It is used to set values in Config
    Attrs dynamically'''

    # pylint: disable=invalid-name

    type: VarType
    var: str

    @staticmethod
    def ENV(var: str, /) -> 'Var':
        '''Shorthand method to define 'Var' object of type 'ENV' '''
        return Var(type=VarType.ENV, var=var)

    @staticmethod
    def SESSION(var: str, /) -> 'Var':
        '''Shorthand method to define 'Var' object of type 'SESSION' '''
        return Var(type=VarType.SESSION, var=var)

    @staticmethod
    def DOC(var: str, /) -> 'Var':
        '''Shorthand method to define 'Var' object of type 'DOC' '''
        return Var(type=VarType.DOC, var=var)

    @staticmethod
    def L10N(var: str, /) -> 'Var':
        '''Shorthand method to define 'Var' object of type 'L10N' '''
        return Var(type=VarType.L10N, var=var)

    @staticmethod
    def CONFIG(var: str, /) -> 'Var':
        '''Shorthand method to define 'Var' object of type 'CONFIG' '''
        return Var(type=VarType.CONFIG, var=var)



'''Provides 'Default' dataclass'''

from dataclasses import dataclass
from typing import Callable, MutableMapping, MutableSequence, Union


@dataclass(kw_only=True)
class Default:
    '''Default dataclass serves role of defining Default Instruction, which
    instructs Nawah to set value for'''

    value: Union[None, str, int, float, bool, MutableMapping, MutableSequence, Callable]

    def __post_init__(self):
        pass



'''Provides 'Job' dataclass'''

from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional, Protocol

if TYPE_CHECKING:
    from croniter import croniter

    from nawah.types import NawahEnv


@dataclass(kw_only=True)
class Job:
    '''Job dataclass serves role of defining items of 'jobs' Config Attr which
    are callabled, called periodically per cron-based schedule'''

    job: 'JobCallable'
    schedule: str
    prevent_disable: bool
    _cron_schedule: 'croniter'
    _next_time: Optional[str] = None
    _disabled: bool = False


class JobCallable(Protocol):
    '''Provides type-hint for 'job' callable of 'Job' '''

    # pylint: disable=too-few-public-methods
    def __call__(
        self,
        env: 'NawahEnv',
    ) -> None:
        ...



'''Provides 'serve_app' Utility'''

import asyncio
import logging
import re
from typing import (TYPE_CHECKING, Any, Callable, MutableMapping,
                    MutableSequence, Optional, Tuple, Union, cast)

from nawah.classes import App, Attr, Default, Var
from nawah.config import Config
from nawah.enums import VarType
from nawah.exceptions import ConfigException

from ._app import _run_app
from ._config import config_app, config_module
from ._val import deep_update, var_value
from ._validate import validate_attr

if TYPE_CHECKING:
    from nawah.classes import (ClientApp, Env, Job, Module, Package, SysDoc,
                               UserSetting)
    from nawah.types import AnalyticsEvents, NawahDoc


logger = logging.getLogger('nawah')


def serve_app(  # pylint: disable=too-many-locals
    name,
    /,
    *,
    version: str,
    anon_token: 'Var',
    admin_password: 'Var',
    debug: Optional[Union[bool, 'Var']] = False,
    port: Optional[int] = None,
    env: Optional[str] = None,
    envs: Optional[MutableMapping[str, 'Env']] = None,
    force_admin_check: Optional[bool] = None,
    vars_types: Optional[MutableMapping[str, 'Attr']] = None,
    vars: Optional[  # pylint: disable=redefined-builtin
        MutableMapping[str, Any]
    ] = None,
    client_apps: Optional[MutableMapping[str, 'ClientApp']] = None,
    analytics_events: Optional['AnalyticsEvents'] = None,
    conn_timeout: Optional[Union[int, 'Var']] = None,
    quota_anon_min: Optional[Union[int, 'Var']] = None,
    quota_auth_min: Optional[Union[int, 'Var']] = None,
    quota_ip_min: Optional[Union[int, 'Var']] = None,
    file_upload_limit: Optional[Union[int, 'Var']] = None,
    file_upload_timeout: Optional[Union[int, 'Var']] = None,
    data_server: Optional[Union[str, 'Var']] = None,
    data_name: Optional[Union[str, 'Var']] = None,
    data_ssl: Optional[Union[bool, 'Var']] = None,
    data_disk_use: Optional[Union[bool, 'Var']] = None,
    cache_server: Optional[Union[str, 'Var']] = None,
    cache_db: Optional[Union[int, 'Var']] = 0,
    cache_username: Optional[Union[str, 'Var']] = None,
    cache_password: Optional[Union[str, 'Var']] = None,
    cache_expiry: Optional[Union[int, 'Var']] = None,
    error_reporting_server: Optional[Union[str, 'Var']] = None,
    locales: Optional[MutableSequence[str]] = None,
    locale: Optional[str] = None,
    admin_doc: Optional['NawahDoc'] = None,
    anon_privileges: Optional[MutableMapping[str, MutableSequence[str]]] = None,
    user_attrs: Optional[MutableMapping[str, 'Attr']] = None,
    user_settings: Optional[MutableMapping[str, 'UserSetting']] = None,
    user_doc_settings: Optional[MutableSequence[str]] = None,
    groups: Optional[MutableSequence[MutableMapping[str, Any]]] = None,
    default_privileges: Optional[MutableMapping[str, MutableSequence[str]]] = None,
    data_indexes: Optional[MutableSequence[MutableMapping[str, Any]]] = None,
    docs: Optional[MutableSequence['SysDoc']] = None,
    jobs: Optional[MutableMapping[str, 'Job']] = None,
    gateways: Optional[MutableMapping[str, Callable]] = None,
    types: Optional[MutableMapping[str, Callable]] = None,
    modules: Optional[MutableSequence[Callable[[], Tuple[str, 'Module']]]] = None,
    packages: Optional[MutableSequence[Callable[[], Tuple[str, 'Package']]]] = None,
):
    '''Returns callable that is used as to serve Nawah App'''

    # pylint: disable=import-outside-toplevel

    from nawah.packages.core import core

    if not re.match(r'^[0-9]+\.[0-9]+\.[0-9]+([ab][0-9]+)?$', version):
        raise ConfigException(
            f'App \'{name}\' defines invalid \'version\' \'{version}\''
        )

    if debug is True:
        logger.setLevel(logging.DEBUG)
        logger.debug('Set logging level to DEBUG (Config.debug==True)')

    if isinstance(debug, Var):
        if debug.type != VarType.ENV:
            raise ConfigException(
                'Only Var objects of type \'ENV\' can be used in Config Attr. Got '
                f'\'{debug.type}\' for debug Config Attr'
            )
        if var_value(debug):
            logger.setLevel(logging.DEBUG)
            logger.debug('Set logging level to DEBUG (os.environ[Config.debug.var])')
            debug = True

    def _():
        return (
            name,
            App(
                version=version,
                debug=debug,
                port=port,
                env=env,
                envs=envs,
                force_admin_check=force_admin_check,
                vars_types=vars_types,
                vars=vars,
                client_apps=client_apps,
                analytics_events=analytics_events,
                conn_timeout=conn_timeout,
                quota_anon_min=quota_anon_min,
                quota_auth_min=quota_auth_min,
                quota_ip_min=quota_ip_min,
                file_upload_limit=file_upload_limit,
                file_upload_timeout=file_upload_timeout,
                data_server=data_server,
                data_name=data_name,
                data_ssl=data_ssl,
                data_disk_use=data_disk_use,
                cache_server=cache_server,
                cache_username=cache_username,
                cache_password=cache_password,
                cache_db=cache_db,
                cache_expiry=cache_expiry,
                error_reporting_server=error_reporting_server,
                locales=locales,
                locale=locale,
                admin_doc=admin_doc,
                admin_password=admin_password,
                anon_token=anon_token,
                anon_privileges=anon_privileges,
                user_attrs=user_attrs,
                user_settings=user_settings,
                user_doc_settings=user_doc_settings,
                groups=groups,
                default_privileges=default_privileges,
                data_indexes=data_indexes,
                docs=docs,
                jobs=jobs,
                gateways=gateways,
                types=types,
                modules=modules,
                packages=packages,
            ),
        )

    config_loaders: MutableSequence[Callable[[], Tuple[str, 'Package']]] = [
        core,
        *(packages or []),
        _,
    ]

    for config_loader in config_loaders:
        config_name, config = config_loader()

        for attr_name in config.__dataclass_fields__:
            attr_val = getattr(config, attr_name, None)

            if attr_val is None:
                continue

            if attr_name in ['api_level', 'version', 'modules', 'packages']:
                continue

            # For vars_types Config Attr, preserve config name for debugging purposes
            if attr_name == 'vars_types':
                for var in attr_val:
                    Config.vars_types[var] = {
                        'package': config_name,
                        'type': attr_val[var],
                    }

            # Otherwise, update Config accroding to attr_val type
            elif isinstance(attr_val, list):
                for j in attr_val:
                    getattr(Config, attr_name).append(j)
                if attr_name == 'locales':
                    Config.locales = list(set(Config.locales))

            elif isinstance(attr_val, dict):
                if not getattr(Config, attr_name):
                    setattr(Config, attr_name, {})
                deep_update(target=getattr(Config, attr_name), new_values=attr_val)

            else:
                setattr(Config, attr_name, attr_val)

        config_modules = []
        if module_loaders := getattr(config, 'modules', None):
            module_loaders = cast(
                MutableSequence[Callable[[], Tuple[str, 'Module']]], module_loaders
            )
            for module_loader in module_loaders:
                module_name, module = module_loader()

                if module_name in Config.modules:
                    raise ConfigException(
                        f'Module \'{module_name}\' exist in current runtime config'
                    )

                config_module(module_name=module_name, module=module)
                config_modules.append(module_name)
                Config.modules[module_name] = module

        if getattr(config, 'packages', None):
            pass

        if isinstance(config, App):
            Config.sys.name = config_name
            Config.sys.version = version
        else:
            Config.sys.packages[config_name] = {
                'package': config,
                'modules': config_modules,
            }

    # [DOC] Update User, Session modules with populated attrs
    Config.modules['user'].attrs.update(Config.user_attrs)
    if sum(1 for attr in Config.user_settings if attr in Config.user_attrs) != 0:
        raise ConfigException(
            'At least one attr from \'user_settings\' is conflicting with an attr from '
            '\'user_attrs\''
        )
    for attr in Config.user_doc_settings:
        Config.modules['user'].attrs[attr] = Config.user_settings[attr].val_type
        Config.modules['user'].attrs[attr].default = Config.user_settings[attr].default
    Config.modules['user'].defaults['locale'] = Default(value=Config.locale)

    for user_attr_name, user_attr in Config.user_attrs.items():
        Config.modules['user'].unique_attrs.append(user_attr_name)
        Config.modules['user'].attrs[f'{user_attr_name}_hash'] = Attr.STR()
        session_auth_doc_args = cast(
            MutableSequence[MutableMapping[str, 'Attr']],
            Config.modules['session'].funcs['auth'].doc_attrs,
        )
        session_auth_doc_args.extend(
            [
                {
                    'hash': Attr.STR(),
                    user_attr_name: user_attr,
                    'groups': Attr.LIST(list=[Attr.ID()]),
                },
                {'hash': Attr.STR(), user_attr_name: user_attr},
            ]
        )

    # [DOC] Attempt to validate all packages required vars (via vars_types Config Attr) are met
    for var_name, var in Config.vars_types.items():
        if var_name not in Config.vars:
            raise ConfigException(
                f'Package \'{var["package"]}\' requires \'{var_name}\' Var, but not found in App '
                'Config'
            )
        try:
            validate_attr(
                mode='create',
                attr_name=var_name,
                attr_type=var['type'],
                attr_val=Config.vars[var_name],
            )
        except Exception as e:
            raise ConfigException(
                f'Package \'{var["package"]}\' requires \'{var_name}\' Var of type '
                f'\'{var["type"].type}\', but validation failed'
            ) from e

    asyncio.run(config_app())
    _run_app()



import asyncio
import datetime
import logging
import time
import traceback

import aiohttp.web

from nawah.config import Config
from nawah.enums import Event

from .._call import call
from ._common_handlers import (_not_allowed_handler, _not_found_handler,
                               _root_handler)
from ._http_handler import _http_handler
from ._utils import _close_session
from ._websocket_handler import _websocket_handler

logger = logging.getLogger('nawah')


async def _jobs_loop():
    while True:
        await asyncio.sleep(60)
        try:
            # Connection Timeout Workflow
            logger.debug('Time to check for sessions!')
            logger.debug(f'Current sessions: {Config.sys.sessions}')
            for session_id, session in Config.sys.sessions.items():
                if 'last_call' not in session.keys():
                    continue
                if datetime.datetime.utcnow() > (
                    session['last_call']
                    + datetime.timedelta(seconds=Config.conn_timeout)
                ):
                    logger.debug(
                        f'Session #\'{session["id"]}\' with REMOTE_ADDR \'{session["REMOTE_ADDR"]}\' HTTP_USER_AGENT: \'{session["HTTP_USER_AGENT"]}\' is idle. Closing'
                    )
                    await _close_session(session_id)
        except Exception:
            logger.error(f'An error occurred. Details: {traceback.format_exc()}')

        try:
            # Calls Quota Workflow - Clean-up Sequence
            logger.debug('Time to check for IPs quotas!')
            del_ip_quota = []
            for ip in Config.sys.ip_quota.keys():
                if (
                    datetime.datetime.utcnow() - Config.sys.ip_quota[ip]['last_check']
                ).seconds > 59:
                    logger.debug(
                        f'IP \'{ip}\' with quota \'{Config.sys.ip_quota[ip]["counter"]}\' is idle. Cleaning-up'
                    )
                    del_ip_quota.append(ip)
            for ip in del_ip_quota:
                del Config.sys.ip_quota[ip]
        except Exception:
            logger.error(f'An error occurred. Details: {traceback.format_exc()}')

        try:
            # Jobs Workflow
            current_time = datetime.datetime.utcnow().isoformat()[:16]
            logger.debug('Time to check for jobs!')
            for job_name in Config.jobs:
                job = Config.jobs[job_name]
                logger.debug(f'Checking: {job_name}')
                if job._disabled:
                    logger.debug('-Job is disabled. Skipping')
                    continue
                # Check if job is scheduled for current_time
                if current_time >= job._next_time:
                    logger.debug('-Job is due, running!')
                    # Update job next_time
                    job._next_time = datetime.datetime.fromtimestamp(
                        job._cron_schedule.get_next(), datetime.timezone.utc
                    ).isoformat()[:16]

                    def job_callback(task: asyncio.Future):
                        if task.done():
                            logger.debug('-Job \'%s\' is done', job_name)

                        if task_exception := task.exception():
                            logger.error(
                                'Job \'%s\' has failed with exception', job_name
                            )
                            logger.error('Exception details:')
                            logger.error(task_exception)
                            if job.prevent_disable:
                                logger.warning(
                                    '-Detected job prevent_disable. Skipping disabling job'
                                )
                            else:
                                logger.warning('-Disabling job')
                                job._disabled = True

                    job_task = asyncio.create_task(job.job(env=Config.sys.env))
                    job_task.add_done_callback(job_callback)

                else:
                    logger.debug('-Not yet due')
        except Exception:
            logger.error(f'An error occurred. Details: {traceback.format_exc()}')

        try:
            logger.debug('Time to check for files timeout!')
            files_results = await call(
                'file/delete',
                skip_events=[Event.PERM],
                env=Config.sys.env,
                query=[
                    {
                        'create_time': {
                            '$lt': (
                                datetime.datetime.utcnow()
                                - datetime.timedelta(seconds=Config.file_upload_timeout)
                            ).isoformat()
                        }
                    }
                ],
            )
            logger.debug('Files timeout results:')
            logger.debug(f'-status: {files_results["status"]}')
            logger.debug(f'-msg: {files_results["msg"]}')
            logger.debug(f'-args.docs: {files_results["args"]["docs"]}')
        except Exception:
            logger.error(f'An error occurred. Details: {traceback.format_exc()}')


def _create_error_middleware(overrides):
    @aiohttp.web.middleware
    async def error_middleware(request, handler):
        try:
            response = await handler(request)
            override = overrides.get(response.status)
            if override:
                return await override(request)
            return response
        except aiohttp.web.HTTPException as ex:
            override = overrides.get(ex.status)
            if override:
                return await override(request)
            raise

    return error_middleware


async def _web_loop():
    # Populate get_routes, post_routes
    get_routes = []
    post_routes = []
    for module_name, module in Config.modules.items():
        for func_name, func in module.funcs.items():
            if func.get_func:
                for get_args_set in func.query_attrs or [{}]:
                    if get_args_set:
                        get_args = f'/{{{"}/{".join(list(get_args_set.keys()))}}}'
                    else:
                        get_args = ''

                    get_routes.append(f'/{module_name}/{func_name}{get_args}')
            elif func.post_func:
                for post_args_set in func.query_attrs or [{}]:
                    if post_args_set:
                        post_args = f'/{{{"}/{".join(list(post_args_set.keys()))}}}'
                    else:
                        post_args = ''

                    post_routes.append(f'/{module_name}/{func_name}{post_args}')

    logger.debug(
        'Loaded modules: %s',
        {module_name: module.attrs for module_name, module in Config.modules.items()},
    )
    logger.debug(
        'Config has attrs: %s',
        {
            k: str(v)
            for k, v in Config.__dict__.items()
            if not isinstance(v, classmethod) and not k.startswith('_')
        },
    )
    logger.debug(f'Generated get_routes: {get_routes}')
    logger.debug(f'Generated post_routes: {post_routes}')

    app = aiohttp.web.Application()
    app.middlewares.append(
        _create_error_middleware(
            {
                404: _not_found_handler,
                405: _not_allowed_handler,
            }
        )
    )
    app.router.add_route('GET', '/', _root_handler)
    app.router.add_route('*', '/ws', _websocket_handler)
    for route in get_routes:
        app.router.add_route('GET', route, _http_handler)
    for route in post_routes:
        app.router.add_route('POST', route, _http_handler)
        app.router.add_route('OPTIONS', route, _http_handler)
    logger.info('Welcome to Nawah')
    runner = aiohttp.web.AppRunner(app)
    await runner.setup()
    site = aiohttp.web.TCPSite(runner, '0.0.0.0', Config.port)
    await site.start()
    logger.info('Serving on 0.0.0.0:%s', Config.port)
    while True:
        await asyncio.sleep(3600)


async def _loop_gather():
    await asyncio.gather(_web_loop(), _jobs_loop())


def _run_app():
    try:
        asyncio.run(_loop_gather())
    except KeyboardInterrupt:
        if time.localtime().tm_hour >= 21 or time.localtime().tm_hour <= 4:
            msg = 'night'
        elif time.localtime().tm_hour >= 18:
            msg = 'evening'
        elif time.localtime().tm_hour >= 12:
            msg = 'afternoon'
        elif time.localtime().tm_hour >= 5:
            msg = 'morning'
        logger.info(f'Have a great {msg}!')
        exit()



'''Provides 'register_module' Utility'''

from typing import (TYPE_CHECKING, Callable, MutableMapping, MutableSequence,
                    Optional, Tuple, Union)

from nawah.classes import Module

if TYPE_CHECKING:
    from nawah.classes import Attr, Counter, Default, Diff, Extn, Func


def register_module(
    name,
    /,
    *,
    funcs: MutableMapping[str, 'Func'],
    collection: str = None,
    attrs: MutableMapping[str, 'Attr'] = None,
    unique_attrs: Optional[MutableSequence[Union[Tuple[str, ...], str]]] = None,
    counters: MutableMapping[str, 'Counter'] = None,
    diff: 'Diff' = None,
    create_draft: bool = False,
    update_draft: bool = False,
    defaults: MutableMapping[str, 'Default'] = None,
    extns: MutableMapping[str, 'Extn'] = None,
    privileges: MutableSequence[str] = None,
    desc: Optional[str] = None,
) -> Callable[[], Tuple[str, 'Module']]:
    '''Returns callable that is used as item for 'modules' Package Config Attr'''

    def _():
        return (
            name,
            Module(
                desc=desc,
                collection=collection,
                attrs=attrs or {},
                funcs=funcs,
                counters=counters or {},
                unique_attrs=unique_attrs or [],
                diff=diff,
                create_draft=create_draft,
                update_draft=update_draft,
                defaults=defaults or {},
                extns=extns or {},
                privileges=privileges
                or ['admin', 'read', 'create', 'update', 'delete'],
            ),
        )

    return _



'''Provides 'call' Utility'''

import asyncio
import datetime
import inspect
import logging
import re
from typing import (TYPE_CHECKING, Any, Awaitable, Callable, MutableMapping,
                    MutableSequence, Union, cast)

from nawah.classes import Query, Var
from nawah.config import Config
from nawah.enums import Event, VarType
from nawah.exceptions import (FuncException, InvalidAttrException,
                              InvalidCallEndpointException,
                              InvalidDocAttrException, InvalidFuncException,
                              InvalidModuleException,
                              InvalidQueryAttrException,
                              MissingDocAttrException,
                              MissingQueryAttrException)

from ._cache import _get_cache, _set_cache
from ._check_permissions import check_permissions as check_permissions_utility
from ._val import camel_to_upper, var_value
from ._validate import validate_attr

if TYPE_CHECKING:
    from nawah.classes import Attr
    from nawah.types import (NawahDoc, NawahEnv, NawahEvents, NawahQuery,
                             Results)


logger = logging.getLogger('nawah')


async def call(
    endpoint: str,
    /,
    *,
    module_name: str = None,
    skip_events: 'NawahEvents' = None,
    env: 'NawahEnv' = None,
    query: 'NawahQuery' = None,
    doc: 'NawahDoc' = None,
    args: MutableMapping[str, Any] = None,
) -> Awaitable['Results']:
    '''Checks validity of an endpoint and calls Nawah Function callable at endpoint,
    returning the couroutine of the callable. If endpoint points to non-existent
    Nawah Function, raises 'InvalidFuncException'.'''

    if not re.match(r'^[a-z_]+\/[a-z_]+$', endpoint):
        raise InvalidCallEndpointException(endpoint=endpoint)

    endpoint_module, endpoint_func = endpoint.split('/')

    try:
        module = Config.modules[endpoint_module]
    except KeyError as e:
        raise InvalidModuleException(module_name=endpoint_module) from e

    try:
        func = module.funcs[endpoint_func]
    except KeyError as e:
        raise InvalidFuncException(
            module_name=endpoint_module, func_name=endpoint_func
        ) from e

    # Set defaults for kwargs
    module_name = module_name or endpoint_module
    skip_events = skip_events or []
    query = query or []
    doc = doc or {}
    env = env or {}

    # Convert query to Query object
    query = Query(query)

    # Check conditions for call checks
    check_permissions = (
        check_attrs_query
    ) = check_attrs_doc = check_cache = check_analytics = True

    check_permissions = Event.PERM not in skip_events
    check_cache = Event.CACHE not in skip_events
    check_attrs_query = check_attrs_doc = Event.ATTRS not in skip_events
    check_analytics = Event.ANALYTICS not in skip_events
    if check_attrs_query:
        check_attrs_query = Event.ATTRS_QUERY not in skip_events
    if check_attrs_doc:
        check_attrs_doc = Event.ATTRS_DOC not in skip_events

    if check_permissions:
        query_mod, doc_mod = check_permissions_utility(func=func, env=env)
        # Use return of check_permissions_utility to update query, doc
        _process_query_mod(
            query=query,
            query_mod=query_mod,
            env=env,
            doc=doc,
        )
        _process_doc_mod(
            doc=doc,
            doc_mod=doc_mod,
            env=env,
            query=query,
        )

    if check_attrs_query:
        func.query_attrs = cast(MutableSequence, func.query_attrs)
        _check_query_attrs(query=query, query_attrs=func.query_attrs)

    if check_attrs_doc:
        func.doc_attrs = cast(MutableSequence, func.doc_attrs)
        _check_doc_attrs(doc=doc, doc_attrs=func.doc_attrs)

    if check_cache:
        cache_key, call_cache = await _get_cache(
            func=func,
            skip_events=skip_events,
            env=env,
            query=query,
        )

        if call_cache:
            return call_cache

    if check_analytics:
        # [TODO] Implement
        pass

    try:
        func_callable = cast(Callable, func.callable)
        kwargs: MutableMapping = {
            'func': func,
            'module_name': module_name,
            'skip_events': skip_events,
            'env': env,
            'query': query,
            'doc': doc,
            'args': args,
        }
        results = await func_callable(
            **{
                param: kwargs[param]
                for param in inspect.signature(func_callable).parameters
                if param in kwargs
            }
        )
    except Exception as e:
        raise FuncException(
            status=getattr(e, 'status', 500),
            msg=e.args[0] if e.args else 'Unexpected error has occurred',
            args={'code': camel_to_upper(e.__class__.__name__)},
        ) from e

    if check_cache and cache_key:
        results['args']['cache_key'] = cache_key
        if 'cache_time' not in results['args']:
            logger.debug(
                'Results generated with \'cache_key\'. Calling \'_set_cache\'.'
            )
            results['args']['cache_time'] = datetime.datetime.utcnow().isoformat()
        asyncio.create_task(_set_cache(func=func, cache_key=cache_key, results=results))

    return results


def _process_query_mod(
    *,
    query: 'Query',
    query_mod: Union[
        MutableMapping[str, Any], MutableSequence[MutableMapping[str, Any]]
    ],
    env: 'NawahEnv',
    doc: 'NawahDoc',
    append_mod: bool = False,
):
    if not query_mod:
        return

    if isinstance(query_mod, list):
        for query_mod_child in query_mod:
            _process_query_mod(query=query, query_mod=query_mod_child, env=env, doc=doc)

    elif isinstance(query_mod, dict):
        for attr_name, attr_val in query_mod.items():
            if isinstance(attr_val, (list, dict)):
                _process_query_mod(query=query, query_mod=attr_val, env=env, doc=doc)
            elif isinstance(attr_val, Var):
                if attr_val.type not in [VarType.SESSION, VarType.CONFIG]:
                    raise Exception(
                        f'\'query_mod\' attr \'{attr_name}\' is of invalid Var type '
                        f'\'{attr_val.type}\''
                    )

                query_mod[attr_name] = var_value(attr_val, env=env, doc=doc)

    if append_mod:
        query.append(query_mod)


def _process_doc_mod(
    *,
    doc: 'NawahDoc',
    doc_mod: Union[MutableMapping[str, Any], MutableSequence[MutableMapping[str, Any]]],
    env: 'NawahEnv',
    query: 'Query',
    append_mod: bool = False,
):
    if not doc_mod:
        return

    if isinstance(doc_mod, list):
        for doc_mod_child in doc_mod:
            _process_doc_mod(doc=doc, doc_mod=doc_mod_child, env=env, query=query)

    elif isinstance(doc_mod, dict):
        for attr_name, attr_val in doc_mod.items():
            if isinstance(attr_val, (list, dict)):
                _process_doc_mod(doc=doc, doc_mod=attr_val, env=env, query=query)
            elif isinstance(attr_val, Var):
                if attr_val.type not in [VarType.SESSION, VarType.CONFIG]:
                    raise Exception(
                        f'\'doc_mod\' attr \'{attr_name}\' is of invalid Var type '
                        f'\'{attr_val.type}\''
                    )

                doc_mod[attr_name] = var_value(attr_val, env=env, doc=doc)

    if append_mod:
        doc_mod = cast(MutableMapping, doc_mod)
        doc.update(doc_mod)


def _check_query_attrs(
    *,
    query: 'Query',
    query_attrs: MutableSequence[MutableMapping[str, 'Attr']],
):
    if not query_attrs:
        return

    for query_attrs_set in query_attrs:
        for attr_name, attr_type in query_attrs_set.items():
            if attr_name not in query:
                raise MissingQueryAttrException(attr_name=attr_name)

            for i, attr_val in enumerate(query[attr_name]):
                try:
                    query[attr_name][i] = validate_attr(
                        mode='create',
                        attr_name=attr_name,
                        attr_type=attr_type,
                        attr_val=attr_val,
                    )
                except InvalidAttrException as e:
                    raise InvalidQueryAttrException(
                        attr_name=attr_name,
                        attr_type=attr_type,
                        val_type=type(attr_val),
                    ) from e


def _check_doc_attrs(
    *,
    doc: 'NawahDoc',
    doc_attrs: MutableSequence[MutableMapping[str, 'Attr']],
):
    if not doc_attrs:
        return

    for doc_attrs_set in doc_attrs:
        for attr_name, attr_type in doc_attrs_set.items():
            if attr_name not in doc:
                raise MissingDocAttrException(attr_name=attr_name)

            try:
                doc[attr_name] = validate_attr(
                    mode='create',
                    attr_name=attr_name,
                    attr_type=attr_type,
                    attr_val=doc[attr_name],
                )
            except InvalidAttrException as e:
                raise InvalidDocAttrException(
                    attr_name=attr_name,
                    attr_type=attr_type,
                    val_type=type(doc[attr_name]),
                ) from e



'''Provides utilities required for configuring modules, packages and app'''

import datetime
import logging
import os
import re
from typing import TYPE_CHECKING, Any, MutableMapping, MutableSequence, cast

import requests
import sentry_sdk
from bson import ObjectId
from croniter import croniter
from passlib.hash import pbkdf2_sha512
from redis import asyncio as aioredis  # type: ignore

from nawah.classes import (Attr, Default, Diff, Extn, Func, Module, SysDoc,
                           UserSetting)
from nawah.config import Config
from nawah.enums import AttrType, Event
from nawah.exceptions import (ConfigException, InvalidAttrTypeArgException,
                              InvalidAttrTypeException)

from ._attr_type import generate_attr_val
from ._call import call
from ._validate._type import validate_type
from ._validate._user_setting import validate_user_setting

if TYPE_CHECKING:
    from nawah.types import NawahDoc

logger = logging.getLogger('nawah')


def extract_attr(*, attrs: MutableMapping[str, 'Attr'], path: str) -> 'Attr':
    '''Extracts attr using its 'path' from 'attrs'. Attr path considers different Attr Types. For
    instance, to get child attr of Attr Type LIST, UNION, use index of child attr, to get 'key',
    'val' attrs of Attr Type KV_DICT use the arg name, to get child attr of Attr Type TYPED_DICT
    use key from 'dict' arg'''

    scope: MutableMapping[str, 'Attr'] | MutableSequence['Attr'] = attrs
    path_part: int | str
    path_parts = path.split('.')
    _iter = enumerate(path_parts)
    for _, path_part in _iter:
        if re.match('^[0-9]+$', cast(str, path_part)):
            path_part = int(path_part)
            path_part_check = path_part in range(len(scope))
        else:
            path_part_check = path_part in scope

        if not path_part_check:
            raise ValueError(
                f'Attr \'{path_part}\' does not exist in scope \'{scope}\''
            )

        scope_attr = cast(MutableMapping[str, 'Attr'], scope)[cast(str, path_part)]

        while scope_attr.type in [AttrType.LIST, AttrType.UNION]:
            try:
                _, path_part = next(_iter)
            except StopIteration:
                return scope_attr

            arg = 'list' if scope_attr.type == AttrType.LIST else 'union'
            scope_attr = scope_attr.args[arg][int(path_part)]

        while scope_attr.type == AttrType.KV_DICT:
            try:
                _, path_part = next(_iter)
            except StopIteration:
                return scope_attr

            scope_attr = scope_attr.args[path_part]

        while scope_attr.type == AttrType.TYPED_DICT:
            try:
                _, path_part = next(_iter)
            except StopIteration:
                return scope_attr

            scope_attr = scope_attr.args['dict'][path_part]


        match scope_attr.type:
            case AttrType.LIST:
                scope = scope_attr.args['list']
            case AttrType.UNION:
                scope = scope_attr.args['union']
            case AttrType.KV_DICT:
                scope = scope_attr.args['val']
            case AttrType.TYPED_DICT:
                scope = scope_attr.args['dict']

    return scope_attr


def config_module(*, module_name: str, module: Module):
    '''Validates args for 'module' and updates values for 'attrs' of 'module' per sets from
    'defaults', 'extns', 'counters' '''

    # pylint: disable=import-outside-toplevel

    import nawah.base as Base

    # Set name of module in Module object for reverse reference
    module.name = module_name

    logger.debug('Attempting to configure module \'%s\'', module_name)

    if not module.funcs:
        raise ConfigException(
            f'Module \'{module_name}\' doesn\'t define any function'
        )

    for func_name, func in module.funcs.items():
        func.name = func_name
        func.module = module
        # Check value type
        if not isinstance(func, Func):
            raise ConfigException(
                f'Invalid Nawah Function for \'{module_name}\'.\'{func_name}\' of type '
                '\'{type(func)}\''
            )
        # If Func has callable defined skip
        if func.callable:
            if not callable(func.callable):
                raise ConfigException(
                    f'Value \'callable\' for Nawah Function \'{func_name}\' of \'{module_name}\' '
                    'is not callable'
                )
            continue

        if func_name.startswith('create_file'):
            func.callable = Base.create_file

        elif func_name.startswith('delete_file'):
            func.callable = Base.delete_file

        elif func_name.startswith('delete_lock'):
            func.callable = Base.delete_lock

        elif func_name.startswith('obtain_lock'):
            func.callable = Base.obtain_lock

        elif func_name.startswith('read_diff'):
            func.callable = Base.read_diff

        elif func_name.startswith('retrieve_file'):
            func.callable = Base.retrieve_file

        elif func_name.startswith('read'):
            func.callable = Base.read

        elif func_name.startswith('create'):
            func.callable = Base.create

        elif func_name.startswith('update'):
            func.callable = Base.update

        elif func_name.startswith('delete'):
            func.callable = Base.delete

        else:
            raise ConfigException(
                f'Can\'t set callable for Nawah Function \'{func_name}\' of'
                f' \'{module_name}\''
            )

    # Check attrs for any invalid type
    for attr in module.attrs:
        try:
            logger.debug(
                'Attempting to validate Attr Type for \'%s\'.\'%s\'',
                module_name,
                attr,
            )
            validate_type(attr_type=module.attrs[attr])
        except InvalidAttrTypeException as e:
            raise ConfigException(
                f'Invalid Attr Type for \'{module_name}\'.\'{attr}\'. Original '
                'validation error: {str(e)}',
            ) from e
        except InvalidAttrTypeArgException as e:
            raise ConfigException(
                f'Invalid Attr Type Arg for \'{module_name}\'.\'{attr}\'. Original '
                'validation error: {str(e)}'
            ) from e

    # Check default for invalid types, update default value
    for attr_name, default in module.defaults.items():
        if not isinstance(default, Default):
            raise ConfigException(
                    f'Invalid Default Instruction for \'{module_name}\'.\'{attr_name}\', of type '
                    '\'{type(default)}\''
            )

        logger.debug(
            'Updating default value for attr \'%s\' to: \'%s\'',
            attr_name,
            module.defaults[attr_name],
        )
        extract_attr(attrs=module.attrs, path=attr_name).default = module.defaults[attr_name]

    # Update extn value
    for attr_name, extn in module.extns.items():
        if not isinstance(extn, Extn):
            raise ConfigException(
                f'Invalid Extension Instruction for \'{module_name}\'.\'{attr_name}\', of type '
                '\'{type(extn)}\''
            )

        logger.debug(
            'Updating \'extn\' for attr \'%s\' to: \'%s\'',
            attr_name,
            module.extns[attr_name],
        )
        extract_attr(attrs=module.attrs, path=attr_name).extn = module.extns[attr_name]

    # Check extns for invalid extended attrs
    for attr_name, extn in module.extns.items():
        if not isinstance(extn, Extn):
            raise ConfigException(
                f'Invalid extns attr \'{attr_name}\' of module \'{module_name}\''
            )

    # Check valid type, value for diff
    if module.diff is not None and not isinstance(module.diff, Diff):
        raise ConfigException(f'Invalid Diff Instruction for module \'{module_name}\'')

    if isinstance(module.diff, Attr):
        module.diff = cast(Attr, module.diff)
        if module.diff.type != AttrType.TYPE:
            raise ConfigException(
                f'Invalid Attr Type for diff of module \'{module_name}\'. Only Attr Type TYPE '
                'is allowed'
            )
        logger.debug(
            'Attempting to validate Attr Type diff of module \'%s\'', module_name
        )
        try:
            validate_type(attr_type=module.diff)
        except InvalidAttrTypeException as e:
            raise ConfigException(
                f'Invalid Attr Type for diff of module \'{module_name}\'. Original validation error: {str(e)}'
            ) from e
        except InvalidAttrTypeArgException as e:
            raise ConfigException(
                f'Invalid Attr Type Arg for diff of module \'{module_name}\'. Original validation error: {str(e)}'
            ) from e

    # [TODO] Re-implement
    # Check valid types, values for create_draft, update_draft
    # for attr in ['create_draft', 'update_draft']:
    #     if type(getattr(self, attr)) not in [bool, Attr]:
    #         raise ConfigException(f'Invalid {attr} for module \'{module_name}\'') from e
    #     if type(getattr(self, attr)) == Attr:
    #         if getattr(self, attr).type != AttrType.TYPE:
    #             raise ConfigException(
    #                 f'Invalid Attr Type for {attr} of module \'{module_name}\'. Only Attr Type TYPE is allowed'
    #             ) from e
    #         logger.debug(
    #             f'Attempting to validate Attr Type {attr} of module \'{module_name}\'.'
    #         )
    #         try:
    #             validate_type(attr_type=getattr(self, attr))
    #         except InvalidAttrTypeException as e:
    #             raise ConfigException(
    #                 f'Invalid Attr Type for {attr} of module \'{module_name}\'. Original validation error: {str(e)}'
    #             ) from e
    #         except InvalidAttrTypeArgException as e:
    #             raise ConfigException(
    #                 f'Invalid Attr Type Arg for {attr} of module \'{module_name}\'. Original validation error: {str(e)}'
    #             ) from e

    logger.debug('Configured module %s', module_name)


async def config_app():

    # [TODO] Add validator for user_attrs, user_doc_settings
    # Check app packages
    if len(Config.sys.dependencies) > 2:
        logger.debug(
            'Found app \'dependencies\'. Attempting to validate all loaded packages are'
            'matching \'dependencies\' value'
        )

        Config.sys.packages['core'] = Config.sys.dependencies['core']
        Config.sys.packages[Config.sys.default_package] = Config.sys.dependencies[
            Config.sys.default_package
        ]

        missing_packages = [
            package
            for package in Config.sys.packages
            if package not in Config.sys.dependencies
        ]
        if missing_packages:
            raise ConfigException(
                'At least one package is missing that is required by app. Missing package[s]:'
                f'\'{", ".join(missing_packages)}\'',
            )

        extra_packages = [
            package
            for package in Config.sys.dependencies
            if package not in Config.sys.packages
        ]
        if extra_packages:
            raise ConfigException(
                'At least one extra package is present in \'packages\' folder that is not'
                f'required by app. Extra package[s]: \'{", ".join(extra_packages)}\''
            )

        # Check for version mismatch
        for package, version in Config.sys.packages.items():
            # Skip core and default_packages
            if package in ['core', Config.sys.default_package]:
                continue
            if version != Config.sys.dependencies[package]:
                raise ConfigException(
                    f'Package \'{package}\' version \'{Config.sys.dependencies[package]}\' is'
                    f'added to app but not matching required version \'{version}\''
                )

    # Check API version
    nawah_level = '.'.join(Config.sys.version.split('.')[0:2])
    for package, api_level in {
        name: config['package'].api_level
        for name, config in Config.sys.packages.items()
    }.items():
        if api_level != nawah_level:
            raise ConfigException(
                f'Nawah framework is on API-level \'{nawah_level}\', but the app package'
                f'\'{package}\' requires API-level \'{api_level}\''
            )
    try:
        versions = (
            (
                requests.get(
                    'https://raw.githubusercontent.com/masaar/nawah_versions/master/versions.txt'
                ).content
            )
            .decode('utf-8')
            .split('\n')
        )
        version_detected = ''
        for version in versions:
            if version.startswith(f'{nawah_level}.'):
                if version_detected and int(version.split('.')[-1]) < int(
                    version_detected.split('.')[-1]
                ):
                    continue
                version_detected = version
        if version_detected and version_detected != Config.sys.version:
            logger.warning(
                f'Your app is using Nawah version \'{Config.sys.version}\' while newer version \'{version_detected}\' of the API-level is available. Please, update'
            )
    except:
        logger.warning(
            'An error occurred while attempting to check for latest update to Nawah. Please, check for updates on your own'
        )

    # Check for jobs
    if Config.jobs:
        # Check jobs schedule validity
        Config.sys.jobs_base = datetime.datetime.utcnow()
        for job_name in Config.jobs:
            job = Config.jobs[job_name]
            if not croniter.is_valid(job.schedule):
                raise ConfigException(
                    f'Job with schedule \'{job_name}\' schedule is invalid'
                )

            job._cron_schedule = croniter(job.schedule, Config.sys.jobs_base)
            job._next_time = datetime.datetime.fromtimestamp(
                job._cron_schedule.get_next(), datetime.timezone.utc
            ).isoformat()[:16]

    # Check for presence of user_auth_attrs
    if not Config.user_attrs:
        raise ConfigException('No \'user_attrs\' are provided')

    # Check for Env Vars
    attrs_defaults = {
        'data_server': 'mongodb://localhost',
        'data_name': 'nawah_data',
        'data_ssl': False,
        'emulate_test': False,
    }
    for attr_name in attrs_defaults:
        attr_val = getattr(Config, attr_name)
        if type(attr_val) == str and attr_val.startswith('$__env.'):
            logger.debug(f'Detected Env Variable for config attr \'{attr_name}\'')
            if not os.getenv(attr_val[7:]):
                logger.warning(
                    f'Couldn\'t read Env Variable for config attr \'{attr_name}\'. Defaulting to \'{attrs_defaults[attr_name]}\''
                )
                setattr(Config, attr_name, attrs_defaults[attr_name])
            else:
                # Set data_ssl to True rather than string Env Variable value
                if attr_name == 'ssl':
                    attr_val = True
                else:
                    attr_val = os.getenv(attr_val[7:])
                logger.warning(
                    f'Setting Env Variable for config attr \'{attr_name}\' to \'{attr_val}\''
                )
                setattr(Config, attr_name, attr_val)

    from nawah import data as Data

    # Create default env dict
    anon_user = _compile_anon_user()
    anon_session = _compile_anon_session()
    anon_session['user'] = anon_user
    Config.sys.conn = Data.create_conn()
    Config.sys.env = {
        'conn': Config.sys.conn,
        'REMOTE_ADDR': '127.0.0.1',
        'HTTP_USER_AGENT': 'Nawah',
        'client_app': '__sys',
        'session': anon_session,
    }

    # Check cache_server Config Attr
    if Config.cache_server:
        logger.debug('Attempting to create connection with \'cache_server\'')
        Config.sys.cache = aioredis.Redis(
            host=Config.cache_server,
            db=Config.cache_db,
            username=Config.cache_username,
            password=Config.cache_password,
        ).json()

    # Check error_reporting_server Config Attr
    if Config.error_reporting_server:
        logger.debug('Attempting to create connection with \'error_reporting_server\'')
        sentry_sdk.init(  # pylint: disable=abstract-class-instantiated
            Config.error_reporting_server,
            traces_sample_rate=1.0,
            with_locals=True,
            release=f'{Config.sys.name}:{Config.sys.version}',
            environment=Config.env,
        )

    # Check test mode
    if Config.test:
        logger.debug('Test mode detected')
        logger.setLevel(logging.DEBUG)
        __location__ = os.path.realpath(os.path.join('.'))
        if not os.path.exists(os.path.join(__location__, 'tests')):
            os.makedirs(os.path.join(__location__, 'tests'))
        for module_name, module in Config.modules.items():
            module_collection = module.collection
            logger.debug(
                f'Updating collection name \'{module_collection}\' of module {module_name}'
            )
            module_collection = module.collection = f'test_{module_collection}'

            logger.debug(f'Flushing test collection \'{module_collection}\'')
            await Data.drop(
                env=Config.sys.env,
                collection_name=module_collection,
            )

    # Test user_settings
    logger.debug('Testing user_settings')
    if Config.user_settings:
        for user_setting in Config.user_settings:
            logger.debug('Testing %s', user_setting)
            if not isinstance(Config.user_settings[user_setting], UserSetting):
                raise ConfigException(
                    'Invalid Config Attr \'user_settings\' with key \'{user_setting}\' of type \'{type(Config.user_settings[user_setting])}\' with required type \'USER_SETTING\''
                )

            # Validate USER_SETTING
            validate_user_setting(user_setting=Config.user_settings[user_setting])

    # Checking users collection
    # [TODO] Updated sequence to handle users
    logger.debug('Testing users collection')
    user_results = await call(
        'user/read',
        # [TODO] Check behaviour with removed Event.ON
        skip_events=[Event.PERM],
        env=Config.sys.env,
        query=[{'_id': 'f00000000000000000000010'}],
    )
    if not user_results['args']['count']:
        logger.debug('ADMIN user not found, creating it')
        # Prepare base ADMIN user doc
        admin_create_doc = {
            '_id': ObjectId('f00000000000000000000010'),
            'name': {Config.locale: '__ADMIN'},
            'groups': [],
            'privileges': {'*': ['*']},
            'locale': Config.locale,
        }
        # Update ADMIN user doc with admin_doc Config Attr
        admin_create_doc.update(Config.admin_doc)

        for auth_attr in Config.user_attrs:
            admin_create_doc[f'{auth_attr}_hash'] = pbkdf2_sha512.using(
                rounds=100000
            ).hash(
                f'{auth_attr}{admin_create_doc[auth_attr]}{Config.admin_password}'
                f'{Config.anon_token}'.encode('utf-8')
            )
        admin_results = await call(
            'user/create',
            skip_events=[Event.PERM],
            env=Config.sys.env,
            doc=admin_create_doc,
        )
        logger.debug('ADMIN user creation results: %s', admin_results)
        if admin_results['status'] != 200:
            raise ConfigException('Config step failed')
    elif not Config.force_admin_check:
        logger.warning(
            'ADMIN user found, skipping check due to force_admin_check Config Attr'
        )
    else:
        logger.warning(
            'ADMIN user found, checking it due to force_admin_check Config Attr'
        )
        admin_doc = user_results['args']['docs'][0]
        admin_doc_update = {}
        for attr_name, attr_val in Config.admin_doc.items():
            if (
                attr_name not in admin_doc
                or not admin_doc[attr_name]
                or attr_val != admin_doc[attr_name]
            ):
                if (
                    isinstance(attr_val, dict)
                    and Config.locale in attr_val
                    and isinstance(admin_doc[attr_name], dict)
                    and (
                        (
                            Config.locale in admin_doc[attr_name]
                            and attr_val[Config.locale]
                            == admin_doc[attr_name][Config.locale]
                        )
                        or (Config.locale not in admin_doc[attr_name])
                    )
                ):
                    continue
                logger.debug(
                    'Detected change in \'admin_doc.%s\' Config Attr', attr_name
                )
                admin_doc_update[attr_name] = attr_val
        for auth_attr in Config.user_attrs:
            auth_attr_hash = pbkdf2_sha512.using(rounds=100000).hash(
                f'{auth_attr}{admin_doc[auth_attr]}{Config.admin_password}'
                f'{Config.anon_token}'.encode('utf-8')
            )
            if (
                f'{auth_attr}_hash' not in admin_doc
                or auth_attr_hash != admin_doc[f'{auth_attr}_hash']
            ):
                logger.debug(f'Detected change in \'admin_password\' Config Attr')
                admin_doc_update[f'{auth_attr}_hash'] = auth_attr_hash
        if len(admin_doc_update.keys()):
            logger.debug(
                f'Attempting to update ADMIN user with doc: \'{admin_doc_update}\''
            )
            admin_results = await call(
                'user/update',
                # [TODO] Check behaviour with removed Event.ON
                # [TODO] Check behaviour with removed Event.PRE
                skip_events=[Event.PERM],
                env=Config.sys.env,
                query=[{'_id': ObjectId('f00000000000000000000010')}],
                doc=admin_doc_update,
            )
            logger.debug(f'ADMIN user update results: {admin_results}')
            if admin_results['status'] != 200:
                raise ConfigException('Config step failed')
        else:
            logger.debug('ADMIN user is up-to-date')

    Config.sys.docs[ObjectId('f00000000000000000000010')] = SysDoc(module='user')

    # Test if ANON user exists
    user_results = await call(
        'user/read',
        # [TODO] Check behaviour with removed Event.ON
        skip_events=[Event.PERM],
        env=Config.sys.env,
        query=[{'_id': 'f00000000000000000000011'}],
    )
    if not user_results['args']['count']:
        logger.debug('ANON user not found, creating it')
        anon_results = await call(
            'user/create',
            # [TODO] Check behaviour with removed Event.ON
            # [TODO] Check behaviour with removed Event.PRE
            skip_events=[Event.PERM],
            env=Config.sys.env,
            doc=_compile_anon_user(),
        )
        logger.debug(f'ANON user creation results: {anon_results}')
        if anon_results['status'] != 200:
            raise ConfigException('Config step failed')
    else:
        logger.debug('ANON user found, checking it')
        anon_doc = _compile_anon_user()
        anon_doc_update = {}
        for attr in Config.user_attrs:
            if attr not in anon_doc or not anon_doc[attr]:
                logger.debug(f'Detected change in \'anon_doc.{attr}\' Config Attr')
                anon_doc_update[attr] = generate_attr_val(attr_type=Config.user_attrs[attr])
        for module in Config.anon_privileges:
            if module not in anon_doc or set(anon_doc[module]) != set(
                Config.anon_privileges[module]
            ):
                logger.debug(f'Detected change in \'anon_privileges\' Config Attr')
                anon_doc_update[f'privileges.{module}'] = Config.anon_privileges[module]
        for auth_attr in Config.user_attrs:
            if (
                f'{auth_attr}_hash' not in anon_doc
                or anon_doc[f'{auth_attr}_hash'] != Config.anon_token
            ):
                logger.debug(f'Detected change in \'anon_token\' Config Attr')
                anon_doc_update[attr] = Config.anon_token
            anon_doc_update[f'{auth_attr}_hash'] = Config.anon_token
        if len(anon_doc_update.keys()):
            logger.debug(
                f'Attempting to update ANON user with doc: \'{anon_doc_update}\''
            )
            anon_results = await call(
                'user/update',
                # [TODO] Check behaviour with removed Event.ON
                # [TODO] Check behaviour with removed Event.PRE
                skip_events=[Event.PERM],
                env=Config.sys.env,
                query=[{'_id': ObjectId('f00000000000000000000011')}],
                doc=anon_doc_update,
            )
            logger.debug(f'ANON user update results: {anon_results}')
            if anon_results['status'] != 200:
                raise ConfigException('Config step failed')
        else:
            logger.debug('ANON user is up-to-date')

    Config.sys.docs[ObjectId('f00000000000000000000011')] = SysDoc(module='user')

    logger.debug('Testing sessions collection')
    # Test if ANON session exists
    session_results = await call(
        'session/read',
        # [TODO] Check behaviour with removed Event.ON
        skip_events=[Event.PERM],
        env=Config.sys.env,
        query=[{'_id': 'f00000000000000000000012'}],
    )
    if not session_results['args']['count']:
        logger.debug('ANON session not found, creating it')
        anon_results = await call(
            'session/create',
            # [TODO] Check behaviour with removed Event.ON
            # [TODO] Check behaviour with removed Event.PRE
            skip_events=[Event.PERM],
            env=Config.sys.env,
            doc=_compile_anon_session(),
        )
        logger.debug(f'ANON session creation results: {anon_results}')
        if anon_results['status'] != 200:
            raise ConfigException('Config step failed')
    Config.sys.docs[ObjectId('f00000000000000000000012')] = SysDoc(module='session')

    logger.debug('Testing groups collection')
    # Test if DEFAULT group exists
    group_results = await call(
        'group/read',
        # [TODO] Check behaviour with removed Event.ON
        skip_events=[Event.PERM],
        env=Config.sys.env,
        query=[{'_id': 'f00000000000000000000013'}],
    )
    if not group_results['args']['count']:
        logger.debug('DEFAULT group not found, creating it')
        group_create_doc = {
            '_id': ObjectId('f00000000000000000000013'),
            'user': ObjectId('f00000000000000000000010'),
            'name': {locale: '__DEFAULT' for locale in Config.locales},
            'bio': {locale: '__DEFAULT' for locale in Config.locales},
            'privileges': Config.default_privileges,
        }
        group_results = await call(
            'group/create',
            # [TODO] Check behaviour with removed Event.ON
            # [TODO] Check behaviour with removed Event.PRE
            skip_events=[Event.PERM],
            env=Config.sys.env,
            doc=group_create_doc,
        )
        logger.debug('DEFAULT group creation results: %s', group_results)
        if group_results['status'] != 200:
            raise ConfigException('Config step failed')
    else:
        logger.debug('DEFAULT group found, checking it')
        default_doc = group_results['args']['docs'][0]
        default_doc_update: MutableMapping[str, Any] = {}
        for module in Config.default_privileges:
            if module not in default_doc['privileges'].keys() or set(
                default_doc['privileges'][module]
            ) != set(Config.default_privileges[module]):
                logger.debug('Detected change in \'default_privileges\' Config Attr')
                default_doc_update[f'privileges.{module}'] = Config.default_privileges[
                    module
                ]
        if len(default_doc_update.keys()):
            logger.debug(
                'Attempting to update DEFAULT group with doc: \'%s\'',
                default_doc_update,
            )
            default_results = await call(
                'group/update',
                # [TODO] Check behaviour with removed Event.ON
                # [TODO] Check behaviour with removed Event.PRE
                skip_events=[Event.PERM],
                env=Config.sys.env,
                query=[{'_id': ObjectId('f00000000000000000000013')}],
                doc=default_doc_update,
            )
            logger.debug(f'DEFAULT group update results: {default_results}')
            if anon_results['status'] != 200:
                raise ConfigException('Config step failed')
        else:
            logger.debug('DEFAULT group is up-to-date')

    Config.sys.docs[ObjectId('f00000000000000000000013')] = SysDoc(module='group')

    # Test app-specific groups
    logger.debug('Testing app-specific groups collection')
    for group in Config.groups:
        group_results = await call(
            'group/read',
            # [TODO] Check behaviour with removed Event.ON
            skip_events=[Event.PERM],
            env=Config.sys.env,
            query=[{'_id': group['_id']}],
        )
        if not group_results['args']['count']:
            logger.debug(
                f'App-specific group with name \'{group["name"]}\' not found, creating it'
            )
            group_results = await call(
                'group/create',
                # [TODO] Check behaviour with removed Event.ON
                # [TODO] Check behaviour with removed Event.PRE
                skip_events=[Event.PERM],
                env=Config.sys.env,
                doc=group,
            )
            logger.debug(
                f'App-specific group with name {group["name"]} creation results: {group_results}'
            )
            if group_results['status'] != 200:
                raise ConfigException('Config step failed')
        else:
            logger.debug(
                f'App-specific group with name \'{group["name"]}\' found, checking it'
            )
            group_doc = group_results['args']['docs'][0]
            group_doc_update = {}
            if 'privileges' in group:
                for module in group['privileges']:
                    if module not in group_doc['privileges'].keys() or set(
                        group_doc['privileges'][module]
                    ) != set(group['privileges'][module]):
                        logger.debug(
                            f'Detected change in \'privileges\' Doc Arg for group with name \'{group["name"]}\''
                        )
                        group_doc_update[f'privileges.{module}'] = group['privileges'][
                            module
                        ]
            if len(group_doc_update.keys()):
                logger.debug(
                    'Attempting to update group with name \'%s\' with doc: \'%s\'',
                    group['name'],
                    group_doc_update,
                )
                group_results = await call(
                    'group/update',
                    # [TODO] Check behaviour with removed Event.ON
                    # [TODO] Check behaviour with removed Event.PRE
                    skip_events=[Event.PERM],
                    env=Config.sys.env,
                    query=[{'_id': group['_id']}],
                    doc=group_doc_update,
                )
                logger.debug(
                    'Group with name \'%s\' update results: %s',
                    group['name'],
                    group_results,
                )
                if group_results['status'] != 200:
                    raise ConfigException('Config step failed')
            else:
                logger.debug('Group with name \'%s\' is up-to-date', group['name'])

        Config.sys.docs[ObjectId(group['_id'])] = SysDoc(module='group')

    # Test app-specific data indexes
    logger.debug('Testing data indexes')
    for index in Config.data_indexes:
        logger.debug('Attempting to create data index: %s', index)
        try:
            Config.sys.conn[Config.data_name][index['collection']].create_index(
                index['index']
            )
        except Exception as e:
            logger.error('Failed to create data index: %s, with error: %s', index, e)
            logger.error('Evaluate error and take action manually')

    logger.debug(
        'Creating \'var\', \'type\', \'user\' data indexes for settings collections'
    )
    Config.sys.conn[Config.data_name]['settings'].create_index([('var', 1)])
    Config.sys.conn[Config.data_name]['settings'].create_index([('type', 1)])
    Config.sys.conn[Config.data_name]['settings'].create_index([('user', 1)])
    logger.debug(
        'Creating \'user\', \'event\', \'subevent\' data indexes for analytics collections'
    )
    Config.sys.conn[Config.data_name]['analytics'].create_index([('user', 1)])
    Config.sys.conn[Config.data_name]['analytics'].create_index([('event', 1)])
    Config.sys.conn[Config.data_name]['analytics'].create_index([('subevent', 1)])
    logger.debug('Creating \'__deleted\' data indexes for all collections')
    for module in Config.modules:
        if Config.modules[module].collection:
            logger.debug(
                'Attempting to create \'__deleted\' data index for collection: %s',
                Config.modules[module].collection,
            )
            Config.sys.conn[Config.data_name][
                Config.modules[module].collection
            ].create_index([('__deleted', 1)])

    # Test app-specific docs
    logger.debug('Testing docs')
    for doc in Config.docs:
        if not isinstance(doc, SysDoc):
            raise ConfigException('Invalid Config Attr \'docs\'')

        doc_results = await call(
            f'{doc.module}/read',
            # [TODO] Check behaviour with removed Event.ON
            # [TODO] Check behaviour with removed Event.PRE
            skip_events=[Event.PERM, Event.ATTRS],
            env=Config.sys.env,
            query=[{doc.key: doc.key_value}],  # type: ignore
        )
        if not doc_results['args']['count']:
            skip_events = [Event.PERM]
            if doc.skip_args is True:
                skip_events.append(Event.ATTRS)
            doc.doc = cast('NawahDoc', doc.doc)
            doc_results = await call(
                f'{doc.module}/create',
                skip_events=skip_events,
                env=Config.sys.env,
                doc=doc.doc,
            )
            logger.debug(
                'App-specific doc with %s \'%s\' of module \'%s\' creation results: %s',
                doc.key,
                doc.key_value,
                doc.module,
                doc_results,
            )
            if doc_results['status'] != 200:
                raise ConfigException('Config step failed')
        Config.sys.docs[ObjectId(doc_results['args']['docs'][0]['_id'])] = SysDoc(
            module=doc.module
        )

    # Check for emulate_test mode
    if Config.emulate_test:
        Config.test = True


def _compile_anon_user():
    anon_doc = {
        '_id': ObjectId('f00000000000000000000011'),
        'name': {Config.locale: '__ANON'},
        'groups': [],
        'privileges': Config.anon_privileges,
        'locale': Config.locale,
    }
    for attr_name, attr_val in Config.user_attrs.items():
        anon_doc[attr_name] = generate_attr_val(attr_type=attr_val)
    for auth_attr in Config.user_attrs:
        anon_doc[f'{auth_attr}_hash'] = Config.anon_token
    return anon_doc


def _compile_anon_session():
    session_doc = {
        '_id': ObjectId('f00000000000000000000012'),
        'user': ObjectId('f00000000000000000000011'),
        'host_add': '127.0.0.1',
        'user_agent': Config.anon_token,
        'timestamp': '1970-01-01T00:00:00',
        'expiry': '1970-01-01T00:00:00',
        'token': Config.anon_token,
        'token_hash': Config.anon_token,
    }
    return session_doc



'''Provides 'Module' dataclass'''

from dataclasses import dataclass
from typing import (TYPE_CHECKING, MutableMapping, MutableSequence, Optional,
                    Tuple, Union)

if TYPE_CHECKING:
    from ._attr import Attr
    from ._counter import Counter
    from ._default import Default
    from ._diff import Diff
    from ._extn import Extn
    from ._func import Func


@dataclass(kw_only=True)
class Module:
    '''Module dataclass serves role of defining a Nawah Datatype along the
    funcs required for it. It is at the centre of Nawah, as in, any call
    in Nawah is directed to a specific Module and fulfilled by one of its
    defined funcs, completed with other configurations defined.'''

    # pylint: disable=too-many-instance-attributes

    desc: Optional[str]
    collection: str
    attrs: MutableMapping[str, 'Attr']
    funcs: MutableMapping[str, 'Func']
    unique_attrs: MutableSequence[Union[Tuple[str], str]]
    counters: MutableMapping[str, 'Counter']
    diff: Optional['Diff']
    create_draft: bool
    update_draft: bool
    defaults: MutableMapping[str, 'Default']
    extns: MutableMapping[str, 'Extn']
    privileges: MutableSequence[str]
    name: Optional[str] = None

    def __post_init__(self):
        # Force setting name to None
        self.name = None



'''Provides 'Diff' dataclass'''

from dataclasses import dataclass
from typing import Callable


@dataclass(kw_only=True)
class Diff:
    '''Diff dataclass serves role of defining Diff Instruction, which defines
    condition for creating diff doc for successful update calls'''

    condition: Callable



'''Provides 'create' Base Function callable'''

import datetime
import logging
from typing import TYPE_CHECKING, Any, MutableSequence, cast

import nawah.data as Data
from nawah.config import Config
from nawah.enums import Event
from nawah.utils import call, expand_val, validate_doc

from .exceptions import (DuplicateUniqueException,
                         UtilityModuleDataCallException)

if TYPE_CHECKING:
    from nawah.types import NawahDoc, NawahEnv, NawahEvents, Results

logger = logging.getLogger('nawah')


async def create(
    *,
    module_name: str,
    skip_events: 'NawahEvents',
    env: 'NawahEnv',
    doc: 'NawahDoc',
) -> 'Results':
    '''Creates doc for a module'''

    module = Config.modules[module_name]

    if not module.collection:
        raise UtilityModuleDataCallException(
            module_name=module_name, func_name='create'
        )

    # Expand dot-notated keys onto dicts
    doc = expand_val(doc=doc)
    # Deleted all extra doc args
    doc = {
        attr: doc[attr]
        for attr in ['_id', *module.attrs]
        if attr in doc and doc[attr] is not None
    }
    # Append host_add, user_agent, create_time, if present in attrs
    if (
        'user' in module.attrs
        and 'host_add' not in doc
        and env['session']
        and Event.ATTRS not in skip_events
    ):
        doc['user'] = env['session']['user']['_id']
    if 'create_time' in module.attrs:
        doc['create_time'] = datetime.datetime.utcnow().isoformat()
    if 'host_add' in module.attrs and 'host_add' not in doc:
        doc['host_add'] = env['REMOTE_ADDR']
    if 'user_agent' in module.attrs and 'user_agent' not in doc:
        doc['user_agent'] = env['HTTP_USER_AGENT']
    if Event.ATTRS not in skip_events:
        # Check presence and validate all attrs in doc args
        validate_doc(
            mode='create',
            doc=doc,
            attrs=module.attrs,
        )
        # Check unique_attrs
        if module.unique_attrs:
            unique_attrs_query: MutableSequence[Any] = [[]]
            for attr in module.unique_attrs:
                if isinstance(attr, str):
                    attr = cast(str, attr)
                    unique_attrs_query[0].append({attr: doc[attr]})
                elif isinstance(attr, tuple):
                    unique_attrs_query[0].append(
                        {child_attr: doc[child_attr] for child_attr in attr}
                    )
                # [TODO] Implement use of single-item dict with LITERAL Attr Type for dynamic unique check based on doc value
            unique_attrs_query.append({'$limit': 1})
            unique_results = await call(
                'base/read',
                module_name=module_name,
                skip_events=[Event.PERM],
                env=env,
                query=unique_attrs_query,
            )
            if unique_results['args']['count']:
                raise DuplicateUniqueException(unique_attrs=module.unique_attrs)
    # Execute Data driver create
    results = await Data.create(env=env, collection_name=module.collection, doc=doc)

    # create soft action is to only return the new created doc _id.
    if Event.SOFT in skip_events:
        read_results = await call(
            'base/read',
            module_name=module_name,
            skip_events=[Event.PERM],
            env=env,
            query=[[{'_id': results['docs'][0]}]],
        )
        results = read_results['args']

    return {'status': 200, 'msg': f'Created {results["count"]} docs.', 'args': results}



'''Provides 'delete' Base Function callable'''

from typing import TYPE_CHECKING

import nawah.data as Data
from nawah.config import Config
from nawah.enums import DeleteStrategy, Event

from .exceptions import UtilityModuleDataCallException

if TYPE_CHECKING:
    from nawah.classes import Query
    from nawah.types import NawahEnv, NawahEvents, ResultsArgs


async def delete(
    *,
    module_name: str,
    skip_events: 'NawahEvents',
    env: 'NawahEnv',
    query: 'Query',
) -> 'ResultsArgs':
    '''Deletes doc from module'''

    module = Config.modules[module_name]

    if not module.collection:
        raise UtilityModuleDataCallException(
            module_name=module_name, func_name='delete'
        )

    # [TODO]: confirm all extns are not linked.
    # [DOC] Pick delete strategy based on skip_events
    strategy = DeleteStrategy.SOFT_SKIP_SYS
    if Event.SOFT not in skip_events and Event.SYS_DOCS in skip_events:
        strategy = DeleteStrategy.SOFT_SYS
    elif Event.SOFT in skip_events:
        if Event.SYS_DOCS not in skip_events:
            strategy = DeleteStrategy.FORCE_SKIP_SYS
        else:
            strategy = DeleteStrategy.FORCE_SYS

    docs_results = results = await Data.read(
        env=env,
        collection_name=module.collection,
        attrs=module.attrs,
        query=query,
        skip_process=True,
    )
    results = await Data.delete(
        env=env,
        collection_name=module.collection,
        docs=[doc['_id'] for doc in docs_results['docs']],
        strategy=strategy,
    )

    return {'status': 200, 'msg': f'Deleted {results["count"]} docs.', 'args': results}



'''Provides 'register_package' Utility'''

import re
from typing import (TYPE_CHECKING, Any, Callable, MutableMapping,
                    MutableSequence, Optional, Tuple, Union)

from nawah.classes import Package

if TYPE_CHECKING:
    from nawah.classes import (Attr, ClientApp, Job, Module, SysDoc,
                               UserSetting, Var)
    from nawah.types import AnalyticsEvents, NawahDoc


def register_package(  # pylint: disable=too-many-locals
    name,
    /,
    *,
    api_level: str,
    version: str,
    vars_types: Optional[MutableMapping[str, 'Attr']] = None,
    vars: Optional[  # pylint: disable=redefined-builtin
        MutableMapping[str, Any]
    ] = None,
    client_apps: Optional[MutableMapping[str, 'ClientApp']] = None,
    analytics_events: Optional['AnalyticsEvents'] = None,
    conn_timeout: Optional[Union[int, 'Var']] = None,
    quota_anon_min: Optional[Union[int, 'Var']] = None,
    quota_auth_min: Optional[Union[int, 'Var']] = None,
    quota_ip_min: Optional[Union[int, 'Var']] = None,
    file_upload_limit: Optional[Union[int, 'Var']] = None,
    file_upload_timeout: Optional[Union[int, 'Var']] = None,
    data_server: Optional[Union[str, 'Var']] = None,
    data_name: Optional[Union[str, 'Var']] = None,
    data_ssl: Optional[Union[bool, 'Var']] = None,
    data_disk_use: Optional[Union[bool, 'Var']] = None,
    cache_server: Optional[Union[str, 'Var']] = None,
    cache_db: Optional[Union[int, 'Var']] = 0,
    cache_username: Optional[Union[str, 'Var']] = None,
    cache_password: Optional[Union[str, 'Var']] = None,
    cache_expiry: Optional[Union[int, 'Var']] = None,
    error_reporting_server: Optional[Union[str, 'Var']] = None,
    locales: Optional[MutableSequence[str]] = None,
    locale: Optional[str] = None,
    admin_doc: Optional['NawahDoc'] = None,
    admin_password: Optional[str] = None,
    anon_token: Optional[str] = None,
    anon_privileges: Optional[MutableMapping[str, MutableSequence[str]]] = None,
    user_attrs: Optional[MutableMapping[str, 'Attr']] = None,
    user_settings: Optional[MutableMapping[str, 'UserSetting']] = None,
    user_doc_settings: Optional[MutableSequence[str]] = None,
    groups: Optional[MutableSequence[MutableMapping[str, Any]]] = None,
    default_privileges: Optional[MutableMapping[str, MutableSequence[str]]] = None,
    data_indexes: Optional[MutableSequence[MutableMapping[str, Any]]] = None,
    docs: Optional[MutableSequence['SysDoc']] = None,
    jobs: Optional[MutableMapping[str, 'Job']] = None,
    gateways: Optional[MutableMapping[str, Callable]] = None,
    types: Optional[MutableMapping[str, Callable]] = None,
    modules: Optional[MutableSequence[Callable[[], Tuple[str, 'Module']]]] = None,
) -> Callable[[], Tuple[str, 'Package']]:
    '''Returns callable that is used as item for 'packages' App Config Attr'''

    if not re.match(r'^[0-9]+\.[0-9]+$', api_level):
        raise Exception(
            f'Package \'{name}\' defines invalid \'api_level\' \'{api_level}\''
        )

    if not re.match(r'^[0-9]+\.[0-9]+\.[0-9]+([ab][0-9]+)?$', version):
        raise Exception(f'Package \'{name}\' defines invalid \'version\' \'{version}\'')

    def _():
        return (
            name,
            Package(
                api_level=api_level,
                version=version,
                vars_types=vars_types,
                vars=vars,
                client_apps=client_apps,
                analytics_events=analytics_events,
                conn_timeout=conn_timeout,
                quota_anon_min=quota_anon_min,
                quota_auth_min=quota_auth_min,
                quota_ip_min=quota_ip_min,
                file_upload_limit=file_upload_limit,
                file_upload_timeout=file_upload_timeout,
                data_server=data_server,
                data_name=data_name,
                data_ssl=data_ssl,
                data_disk_use=data_disk_use,
                cache_server=cache_server,
                cache_username=cache_username,
                cache_password=cache_password,
                cache_db=cache_db,
                cache_expiry=cache_expiry,
                error_reporting_server=error_reporting_server,
                locales=locales,
                locale=locale,
                admin_doc=admin_doc,
                admin_password=admin_password,
                anon_token=anon_token,
                anon_privileges=anon_privileges,
                user_attrs=user_attrs,
                user_settings=user_settings,
                user_doc_settings=user_doc_settings,
                groups=groups,
                default_privileges=default_privileges,
                data_indexes=data_indexes,
                docs=docs,
                jobs=jobs,
                gateways=gateways,
                types=types,
                modules=modules,
            ),
        )

    return _



'''Provides 'Cache' dataclass'''

from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional, Protocol

from nawah.config import Config

if TYPE_CHECKING:
    from nawah.classes import Query
    from nawah.types import NawahEnv, NawahEvents


@dataclass(kw_only=True)
class Cache:
    '''Cache dataclass serves role of defining Cache Instruction, which
    instructs Nawah of when to cache call results'''

    channel: str
    condition: 'CacheCondition'
    user_scoped: bool
    period: Optional[int] = None

    def __post_init__(self):
        Config.sys.cache_channels.add(self.channel)


class CacheCondition(Protocol):
    '''Provides type-hint for 'condition' callable of 'Cache' '''

    # pylint: disable=too-few-public-methods

    def __call__(
        self,
        skip_events: 'NawahEvents',
        env: 'NawahEnv',
        query: 'Query',
    ) -> bool:
        ...



'''Provides 'Counter' dataclass'''

from dataclasses import dataclass
from typing import Optional, Protocol


@dataclass(kw_only=True)
class Counter:
    '''Counter dataclass serves role as Counter Value Instruction'''

    pattern_formatter: 'CounterPatternFormatter'
    counter: Optional[str] = None


class CounterPatternFormatter(Protocol):
    '''Provides type-hint for 'pattern_formatter' callable of 'Counter' '''

    # pylint: disable=too-few-public-methods

    def __call__(
        self,
        counter_value: str,
    ) -> str:
        ...



import argparse
import logging
import os
import re

logger = logging.getLogger('nawah')

NAME_PATTERN = r'^[a-z][a-z0-9_]+$'


def create(args: argparse.Namespace):
    '''Provides create command functionality to Nawah CLI'''

    if not re.match(r'^[a-z][a-z0-9_]+$', args.name):
        raise Exception(
            'Value for \'name\' CLI Arg is invalid. Name should have only small letters, '
            'numbers, and underscores.'
        )

    component = args.component
    path = os.path.realpath(args.path)

    logger.info(
        'Checking possibility to create component \'%s\', at \'%s\'', component, path
    )



'''Provides 'core' Nawah Package'''

from nawah.utils import register_package

from .base import base
from .file import file
from .group import group
from .session import session
from .setting import setting
from .user import user

core = register_package(
    'core',
    api_level='2.0',
    version='2.0.0',
    modules=[base, user, group, session, setting, file],
)



'''Provides 'setting' Module Functions callables'''

from nawah.config import Config
from nawah.enums import Event
from nawah.exceptions import FuncException, InvalidAttrException
from nawah.utils import call, decode_attr_type, validate_doc


async def _create(env, doc):
    create_results = await call('base/create', module_name='setting', env=env, doc=doc)

    if create_results['status'] != 200:
        return create_results

    if doc['type'] in ['user', 'user_sys']:
        if (
            doc['user'] == env['session']['user']['_id']
            and doc['var'] in Config.user_doc_settings
        ):
            env['session']['user'][doc['var']] = doc['val']

    return create_results


async def _update(env, query, doc):
    for attr in doc.keys():
        if attr == 'val' or attr.startswith('val.'):
            val_attr = attr
            break
    else:
        raise FuncException(
            status=400,
            msg='Could not match doc with any of the required doc_args. Failed sets: '
            '[\'val\': Missing]',
            args={'code': 'INVALID_DOC'},
        )

    setting_results = await call(
        'setting/read',
        skip_events=[Event.PERM],
        env=env,
        query=query,
    )
    if not setting_results['args']['count']:
        raise FuncException(
            status=400, msg='Invalid Setting doc', args={'code': 'INVALID_SETTING'}
        )

    setting = setting_results['args']['docs'][0]
    # [DOC] Attempt to validate val against Setting val_type
    try:
        exception_raised: Exception = None
        setting_val_type = decode_attr_type(encoded_attr_type=setting['val_type'])
        await validate_doc(
            mode='update',
            doc=doc,
            attrs={'val': setting_val_type},
        )
    except InvalidAttrException as e:
        exception_raised = e

    if exception_raised or doc[val_attr] is None:
        raise FuncException(
            status=400,
            msg=f'Invalid value for for Setting doc of type \'{type(doc[val_attr])}\' with '
            f'required type \'{setting["val_type"]}\'',
            args={'code': 'INVALID_ATTR'},
        )

    create_results = await call('base/create', module_name='setting', env=env, doc=doc)

    if create_results['status'] != 200:
        return create_results

    if doc['type'] in ['user', 'user_sys']:
        if (
            doc['user'] == env['session']['user']['_id']
            and doc['var'] in Config.user_doc_settings
        ):
            env['session']['user'][doc['var']] = doc['val']

    return create_results
